#!/usr/bin/env python
# -*- coding:utf-8 -*-


##############################################################################
## license :
##============================================================================
##
## File :        Smaract_mcs.py
##
## Project :    SmaractMCS
##
## This file is part of Tango device class.
##
## Tango is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Tango is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Tango.  If not, see <http://www.gnu.org/licenses/>.
##
##
## $Author :      sci.comp@elettra.eu
##
## $Revision :    $
##
## $Date :        2015
##
## $HeadUrl :     $
##============================================================================
##            This file was generated by POGO 
##    (Program Obviously used to Generate tango Object)
##        (c) - Software Engineering Group - ESRF
##############################################################################

""""""

__all__ = ["Smaract_mcs", "Smaract_mcsClass", "main"]

__docformat__ = 'restructuredtext'

import PyTango
import sys
# Add additional import
import time
import math

SENSOR_TYPE_S = 1       # linear positioner with nano sensor
SENSOR_TYPE_SR = 2      # rotary positioner with nano sensor
SENSOR_TYPE_SR20 = 8    # rotary positioner with nano sensor
SENSOR_TYPE_G605S = 16  # goniometer with nano sensor (60.5mm radius)
SENSOR_TYPE_G775S = 17  # goniometer with nano sensor (77.5mm radius)
SENSOR_TYPE_SD = 21     # extended scanning Range linear positioner with nano sensor
SENSOR_TYPE_SR2 = 23    # rotary positioner with nano sensor for high applied masses
SENSOR_TYPE_G605DS = 48 # goniometer with nano sensor (60.5mm radius), large actuator
SENSOR_TYPE_G775DS = 49 # goniometer with nano sensor (77.5mm radius), large actuator

BEGIN_OF_LINE = ':'
END_OF_LINE = '\012'

## Device States Description

class Smaract_mcs (PyTango.Device_4Impl):

    #--------- Add you global variables here --------------------------

    def __init__(self,cl, name):
        PyTango.Device_4Impl.__init__(self,cl,name)
        self.debug_stream("In __init__()")
        Smaract_mcs.init_device(self)


    def delete_device(self):
        self.debug_stream("In delete_device()")
        self.stop_polling()
        del self.socket
        del self.autoproxy


    def init_device(self):
        self.debug_stream("In init_device()")
        self.get_device_properties(self.get_device_class())
        # self.attr_Steps_read = 1
        self.attr_Position_read = 0.0
        self.attr_Acceleration_read = 0.0
        self.attr_Velocity_read = 0.0
        self.attr_MotorTitle_read = ""
        self.attr_DefinePosition_read = 0.0
        self.attr_Home_position_read = 0.0
        self.attr_StepSize_read = 0.0
        self.attr_InvertUpDown_read = False
        self.attr_ClosedLoop_read = False
        self.attr_NextPosition_read = 0.0
        self.attr_PreviousPosition_read = 0.0
        self.attr_Error_read = ''
        self.attr_RequireGo_read = False
        self.attr_HighLimitPosition_read = 0.0
        self.attr_LowLimitPosition_read = 0.0
        self.attr_PositionBackup_read = 0.0
        self.attr_SensorEnabled_read = 0
        self.attr_SensorStatus_read = ''
        self.attr_PhysicalPositionKnown_read = False
        self.attr_BiDirReferencing_read = False
        self.set_state(PyTango.DevState.INIT)
        self.status_string = "Initializing device..."
        self.hold_time_ms = 2000
        self.error_msg_timer = 0.0
        self.safe_direction = -1

        self.linear_sensors = (SENSOR_TYPE_S, SENSOR_TYPE_SD)
        self.rotary_sensors = (SENSOR_TYPE_SR, SENSOR_TYPE_SR20,
            SENSOR_TYPE_G605S, SENSOR_TYPE_G775S, SENSOR_TYPE_SR2,
            SENSOR_TYPE_G605DS, SENSOR_TYPE_G775DS)

        self.autoproxy = PyTango.DeviceProxy(self.get_name())

        try:
            # socket device connected to a SmarAct MCS Controller
            self.socket = PyTango.DeviceProxy(self.SocketDevice)
            if self.socket.state() != PyTango.DevState.ON:
                # Try to reconnect once...
                self.socket.Reconnect()
                time.sleep(0.5)
                if self.socket.state() != PyTango.DevState.ON:
                    self.set_state(PyTango.DevState.FAULT)
                    self.status_string = "Socket device not in ON state. INIT failed."
                    return

        except PyTango.DevFailed:
            exctype, value = sys.exc_info()[:2]
            print "Failed with exception ! ", exctype
            for err in value:
                print " reason", err.reason
                print " description", err.desc
                print " origin", err.origin
                print " severity", err.severity

            self.set_state(PyTango.DevState.FAULT)
            self.status_string = "INIT failed, Socket proxy failed."
            self.debug_stream(self.status_string)
            return

        self.debug_stream("Channel: "+str(self.ChannelIndex))
        self.debug_stream("SensorType: "+str(self.SensorType))

        time.sleep(0.1)
        # Set Communication Mode to Sync (0) (Alt. Async(1))
        resp = self.socket_write_and_read("GCM")
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            self.set_state(PyTango.DevState.FAULT)
            PyTango.Except.throw_exception("Set Communication Mode to Sync failed.", self.status_string, "init_device()")
        else:
            try:
                # Expected repy is CM0 or CM1
                comm_mode = int(resp[-1])
            except Exception, e:
                PyTango.Except.throw_exception("Set Communication Mode to Sync failed.", "Unexpected reply format", "init_device()")

        self.debug_stream('Communication mode:'+str(comm_mode))
        if comm_mode != 0:
            self.debug_stream('Communication mode is not 0 (Sync), set it to sync.')
            time.sleep(0.1)
            # Set Communication Mode to Sync (0) (Alt. Async(1))
            resp = self.socket_write_and_read("SCM0")
            if resp == 0:
                self.debug_stream(self.attr_Error_read)
                self.set_state(PyTango.DevState.FAULT)
                PyTango.Except.throw_exception("Set Communication Mode to Sync failed.", self.status_string, "init_device()")
            else:
                self.debug_stream("init_device(): Sync Communication Mode set (for all channels).")

        time.sleep(0.1)
        # Check Sensor type:
        cmd_str = "GST"+str(self.ChannelIndex)
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            # raise Exception("Incorrect request, check socket status.")
            self.set_state(PyTango.DevState.FAULT)
            self.debug_stream(self.status_string)
            PyTango.Except.throw_exception("Get Sensor Type failed.", self.status_string, "init_device()")
        else:
            try:
                stype = int(str.split(resp, ',')[1])
            except Exception, e:
                PyTango.Except.throw_exception("Get Sensor Type failed.", "Unexpected reply format", "init_device()")

        self.debug_stream("Sensor Type:"+str(stype))
        if stype != self.SensorType:
            time.sleep(0.1)
            # Set Sensor type (property SensorType)
            if self.SensorType in self.linear_sensors + self.rotary_sensors:
                cmd_str = "SST" + str(self.ChannelIndex) + "," + str(self.SensorType)
            else:
                self.set_state(PyTango.DevState.FAULT)
                self.status_string = "INIT failed, Invalid SensorType code for the current setup."
                self.debug_stream(self.status_string)
                PyTango.Except.throw_exception("Invalid Sensor Type.", self.status_string, "init_device()")

            resp = self.socket_write_and_read(cmd_str)
            if resp == 0:
                # raise Exception("Incorrect request, check socket status.")
                self.set_state(PyTango.DevState.FAULT)
                self.debug_stream(self.status_string)
                PyTango.Except.throw_exception("Set Sensor Type failed.", self.status_string, "init_device()")

        if self.SensorType in self.rotary_sensors:
            self.SensorTypeRotary = True
            self.debug_stream('Sensor is a Rotary type.\n')
        else:
            self.SensorTypeRotary = False
            self.debug_stream('Sensor is a Linear type.\n')

        # Rotary sensors unit move is in microdegrees. We shall use degrees
        # Linear sensors unit move is in nanometers. We shall use milimeters.
        self.SensorUnitScaleFactor = 1e6

        time.sleep(0.2)
        # Check Power Save Mode
        resp = self.socket_write_and_read("GSE")
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            self.set_state(PyTango.DevState.FAULT)
            PyTango.Except.throw_exception("Get Sensors Enabled failed.", self.status_string, "init_device()")
        else:
            try:
                # expected reply SE<0/1/2>
                sensor_enabled = int(resp[-1])
            except Exception, e:
                PyTango.Except.throw_exception("Get Sensors Enabled failed.", "Unexpected reply format", "init_device()")

        self.debug_stream('Sensor enabled: '+str(sensor_enabled))
        if sensor_enabled != 2:
            time.sleep(0.1)
            # Enable sensor, set to Power Save mode.
            if self.socket_write_and_read("SSE2") == 0:
                self.debug_stream(self.attr_Error_read)
                self.set_state(PyTango.DevState.FAULT)
                PyTango.Except.throw_exception("Enable Sensor in Power Save Mode failed", self.status_string, "init_device()")
            else:
                self.debug_stream("init_device(): Sensor enabled in power save mode (for all channels).")

        # Check various startup attributes:

        # Acceleration:
        time.sleep(0.1)
        cmd_str = "GCLA"+str(self.ChannelIndex)
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Read Acceleration Failed", self.status_string, "init_device()")
        else:
            # print 'Reply:',resp
            try:
                read_acc = float(str.split(resp,',')[1])
                self.attr_Acceleration_read = read_acc / self.SensorUnitScaleFactor
                self.debug_stream("Acceleration as set on controller: "+str(self.attr_Acceleration_read))
            except Exception, e:
                PyTango.Except.throw_exception("Read Acceleration Failed", "Unexpected reply format", "init_device()")

        # Velocity
        time.sleep(0.1)
        cmd_str = "GCLS"+str(self.ChannelIndex)
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Read Velocity Failed", self.status_string, "init_device()")
        else:
            # print 'Reply:',resp
            try:
                read_speed = float(str.split(resp,',')[1])
                self.attr_Velocity_read = read_speed / self.SensorUnitScaleFactor
                self.debug_stream("Velocity as set on controller"+str(self.attr_Velocity_read))
            except Exception, e:
                PyTango.Except.throw_exception("Read Velocity Failed", "Unexpected reply format", "init_device()")

        # If the max speed is not set, try to set it now.
        if self.attr_Velocity_read == 0.0:
            time.sleep(0.1)
            # Set Closed Loop move Speed to 1/20th max speed.
            # Movement speed is given in nm/s for linear positioners and in micro-degrees/s for rotary positioners.
            # The valid range is 0 .. 100,000,000.
            # A value of 0 (default) deactivates the speed control feature.
            #
            # 10% max speed = 10 mm/s || 10 degrees/s
            cmd = "SCLS" + str(self.ChannelIndex) + ",5000000"
            if self.socket_write_and_read(cmd) == 0:
                self.debug_stream(self.attr_Error_read)
                self.set_state(PyTango.DevState.FAULT)
                PyTango.Except.throw_exception("Set Closed Loop Speed failed.", self.status_string, "init_device()")
                # If the error is no sensor attached, you might have to reset the controller.
            else:
                self.debug_stream("init_device(): Set Closed Loop Speed executed.")

        # Direction inversion and scale shift
        time.sleep(0.1)
        self.attr_InvertUpDown_read, self.attr_DefinePosition_read = self.read_scale()

        time.sleep(0.1)
        # Get physical position known (1 - yes, 0 - no)
        cmd = "GPPK" + str(self.ChannelIndex)
        resp = self.socket_write_and_read(cmd)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            self.set_state(PyTango.DevState.FAULT)
            PyTango.Except.throw_exception("Get Physical Position Known failed", self.status_string, "init_device()")
        else:
            if int(str.split(resp, ',')[1]):
                # Expected result is PPK,<val> 1 = known, 0 = unknown
                self.debug_stream("init_device(): Physical Position Known.\n")
                self.attr_PhysicalPositionKnown_read = True
                self.set_state(PyTango.DevState.ON)
                self.status_string = "Device ready."
            else:
                self.debug_stream("init_device(): Physical Position Unnown!\n")
                self.attr_PhysicalPositionKnown_read = False
                self.set_state(PyTango.DevState.ALARM)
                self.status_string = "Physical Position Unknown!"

        self.set_status(self.status_string)

        if self.attr_PhysicalPositionKnown_read:
            self.attr_LowLimitPosition_read, self.attr_HighLimitPosition_read = self.read_limits()
            self.attr_Position_read = self.read_logical_position()
        
    # init_device() complete.


    #-----------------------------------------------------------------------------
    #    Always executed hook
    #-----------------------------------------------------------------------------
    def always_executed_hook(self):
        # self.debug_stream("In always_excuted_hook()")
        if self.socket.state() != PyTango.DevState.ON:
            self.set_state(PyTango.DevState.FAULT)
            self.status_string = "Socket device not in ON state, check connection."
            self.socket_reconnect()
        elif not self.attr_PhysicalPositionKnown_read:
            self.set_state(PyTango.DevState.ALARM)
            self.status_string = "Physical position unknown."            
        else:
            cmd_str = "GS"+str(self.ChannelIndex)
            resp = self.socket_write_and_read(cmd_str)
            if resp == 0:
                self.debug_stream(self.attr_Error_read)
                self.set_state(PyTango.DevState.ALARM)
            else:
                try:
                    status_code = int(str.split(resp,',')[1])
                    self.status_string = self.get_movement_status(status_code)
                    if status_code in [0,3,5]:
                        self.set_state(PyTango.DevState.ON)
                    elif status_code in [1,2,4,6,7]:
                        self.set_state(PyTango.DevState.MOVING)
                    elif status_code == 9:
                        self.set_state(PyTango.DevState.ALARM)
                    else:
                        self.set_state(PyTango.DevState.UNKNOWN)
                except Exception, e:
                    self.set_state(PyTango.DevState.ALARM)
                    self.status_string = "Failed parsing the controller status."
        self.set_status(self.status_string)


    #-----------------------------------------------------------------------------
    #    Utils
    #-----------------------------------------------------------------------------

    def socket_reconnect(self):
        ''' Try to reconnect the socket. ''' 
        try:
            self.socket.Reconnect()
        except Exception, e:
            print "Socket connection raised exception.\n"
            self.status_string = "Socket Failed: "+str(e)

    #------------------------------------------------------------------
    #    Read/write socket utility
    #        Returns a string in case of a proper reply,
    #        0 for no reply or failure.
    #------------------------------------------------------------------
    def socket_write_and_read(self, argin):
        # print "In ", self.get_name(), "::socket_write_and_read()"
        # self.debug_stream("argin: " + str(argin))
        mcs_cmd = BEGIN_OF_LINE + str(argin) + END_OF_LINE

        attempts = 0
        max_attempts = 2 # In case of a socket exception, retry once.
        while (attempts < max_attempts):
            try:
                argout = self.socket.WriteAndRead(mcs_cmd)
                if len(argout) == 0:
                    self.status_string = "Cmd failed, no reply." # We ARE in the 'Sync' mode.
                    self.debug_stream(self.status_string)
                    return 0

                # Strip the BEGIN_OF_LINE and END_OF_LINE characters from the response.
                argout = argout[1:-1]
                # self.debug_stream("argout: " + str(argout))

                if argout.startswith('E'):
                    try:
                        err_code = int(str.split(argout, ',')[1])
                    except Exception, e:
                        print e
                        return 0
                    if err_code == 0:
                        # self.status_string = "Cmd " + str(argin) + " successfully executed.\n"
                        return argout
                    else:
                        err_type, err_desc = self.get_error_msg(err_code)
                        self.status_string = "Cmd failed with " + err_type
                        self.attr_Error_read = err_type + " : " + err_desc
                        self.debug_stream(self.status_string)
                        self.error_msg_timer = time.time()
                        return 0
                else:
                    # self.status_string = "Cmd " + str(argin) + " successfully executed.\n"
                    return argout
            except Exception, e:
                print "Socket connection raised exception.\n"
                attempts += 1
                self.status_string = "Socket Failed: "+str(e)
                time.sleep(0.1)

        # Socket problem persists, set state to FAULT
        self.set_state(PyTango.DevState.FAULT)
        return 0

    #------------------------------------------------------------------
    #    Get MCS error type utility
    #        argin = error code (int)
    #        Returns a string with error type and description,
    #        None if error code is not valid.
    #------------------------------------------------------------------
    def get_error_msg(self, argin):

        error_type = {
            0 : "No Error",
            1 : "Syntax Error",
            2 : "Invalid Command Error",
            3 : "Overflow Error",
            4 : "Parse Error",
            5 : "Too Few Parameters Error",
            6 : "Too Many Parameters Error",
            7 : "Invalid Parameter Error",
            8 : "Wrong Mode Error",
            129 : "No Sensor Present Error",
            140 : "Sensor Disabled Error",
            141 : "Command Overridden Error",
            142 : "End Stop Reached Error",
            143 : "Wrong Sensor Type Error",
            144 : "Could Not Find Reference Mark Error",
            145 : "Wrong End Effector Type Error",
            146 : "Movement Locked Error",
            147 : "Range Limit Reached Error",
            148 : "Physical Position Unknown Error",
            150 : "Command Not Processable Error",
            151 : "Waiting For Trigger Error",
            152 : "Command Not Triggerable Error",
            153 : "Command Queue Full Error",
            154 : "Invalid Component Error",
            155 : "Invalid Sub Component Error",
            156 : "Invalid Property Error",
            157 : "Permission Denied Error"
        }

        error_desc = {
            0 : "This indicates that no error occurred and therefore corresponds to an acknowledge.",
            1 : "The command could not be processed due to a syntactical error.",
            2 : "The command given is not known to the system.",
            3 : "This error occurs if a parameter given is too large and therefore cannot be processed.",
            4 : "The command could not be processed due to a parse error.",
            5 : "The specified command requires more parameters in order to be executed.",
            6 : "There were too many parameters given for the specified command.",
            7 : "A parameter given exceeds the valid range. Please see the command description "\
                "for valid ranges of the parameters.",
            8 : "This error is generated if the specified command is not available in the "\
                "current communication mode. "\
                "For example, the SRC  command is not executable in synchronous mode.",
            129 : "This error occurs if a command was given that requires sensor feedback, "\
                "but the addressed positioner has none attached.",
            140 : "This error occurs if a command was given that requires sensor feedback, "\
                "but the sensor of the addressed positioner is disabled (see SSE  command).",
            141 : "This error is only generated in the asynchronous communication mode. "\
                "When the software commands a movement which is then interrupted by the "\
                "Hand Control Module, an error of this type is generated.",
            142 : "This error is generated in asynchronous mode if the target position of a "\
                "closed-loop command could not be reached, because a mechanical end stop was detected. "\
                "After this error the positioner will have a movement status code of 0 (stopped).",
            143 : "This error occurs if a closed-loop command does not match the sensor type that "\
                "is currently configured for the addressed channel. For example, issuing a GP  "\
                "command while the targeted channel is configured as rotary will lead to this error.",
            144 : "This error is generated in asynchronous mode (see SCM ) if the search "\
                "for a reference mark was aborted.",
            145 : "This error occurs if a command does not match the end effector type that is currently "\
                "configured for the addressed channel. For example, sending GF  while the targeted channel "\
                "is configured for a gripper will lead to this error.",
            146 : "This error occurs if a movement command is issued while the system is in the locked state.",
            147 : "If a range limit is defined (SPL  or SAL ) and the positioner is about to move beyond "\
                "this limit, then the positioner will stop and report this error (only in asynchronous mode, "\
                "see SCM ). After this error the positioner will have status code of 0 (stopped).",
            148 : "A range limit is only allowed to be defined if the positioner knows its physical position. "\
                "If this is not the case, the commands SPL  and SAL  will return this error code.",
            150 : "This error is generated if a command is sent to a channel when it is in a state where the "\
                "command cannot be processed. For example, to change the sensor type of a channel the addressed "\
                "channel must be completely stopped. In this case send a stop command before changing the type.",
            151 : "If there is at least one command queued in the command queue then you may only append more "\
                "commands (if the queue is not full), but you may not issue movement commands for immediate "\
                "execution. Doing so will generate this error. ",
            152 : "After sending a ATC  command you are required to issue a movement command that is to be "\
                "triggered by the given event source. Commands that cannot be triggered will generate this error.",
            153 : "This error is generated if you attempt to append more commands to the command queue, "\
                "but the queue cannot hold anymore commands. The queue capacity may be read out with a "\
                "get channel property command (GCP).",
            154 : "Indicates that a component (e.g. SCP) was selected that does not exist.",
            155 : "Indicates that a sub component (e.g. SCP) was selected that does not exist.",
            156 : "Indicates that a property (e.g. SCP) was selected that does not exist.",
            157 : "This error is generated when you call a functionality which is not unlocked "\
                "for the system (e.g. Low Vibration Mode)."
        }

        if argin in error_type:
            return error_type[argin], error_desc[argin]
        else:
            return None


    #------------------------------------------------------------------
    #    Get movement status utility
    #        argin = movement status code (int), as returned by the GS command.
    #        Returns a string with error type and description,
    #        None if error code is not valid.
    #------------------------------------------------------------------
    def get_movement_status(self, argin):

        channel_status = {
            0 : "Stopped", # " - The positioner is currently not performing active movement (see S command).",
            1 : "Stepping", # " - The positioner is performing an open-loop movement (see MST command).",
            2 : "Scanning", # " - The positioner is performing a scanning movement (see commands MSCA and MSCR).",
            3 : "Holding", # " - The positioner is holding its current target position resp. angle (see closed-loop"\
                # "commands, e.g. MPA or MAA) or is holding the reference mark (see FRM command).",
            4 : "Targeting", # " - The positioner is performing a closed-loop movement "\
                # "(see closed-loop commands, e.g. MPA or MAA).",
            5 : "Move Delay", # " - The positioner is currently waiting for the sensors to power up before executing the"\
                # "movement command. This status may be returned if the the sensors are operated in power save mode.",
            6 : "Calibrating", # " - The positioner is busy calibrating its sensor (see commands CS and SZF).",
            7 : "Finding Reference Mark", # " - The positioner is moving to find the reference mark (see FRM command).",
            9 : "Locked", # " - An emergency stop has occurred and further movements are not allowed (see SESMcommand)."
        }

        if argin in channel_status:
            return channel_status[argin]
        else:
            return None


    def read_scale(self):
        """ Returns scale inversion (True/False) and scale shift (mm/deg). """
        cmd_str = "GSC"+str(self.ChannelIndex)
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Get Scale failed", self.status_string, "read_scale()")
        else:
            # print 'Reply:',resp
            try:
                scale_shift = float(str.split(resp, ',')[1]) / self.SensorUnitScaleFactor
                inverted = int(str.split(resp, ',')[2])
                if inverted:
                    up_down = True
                else:
                    up_down = False

                if self.SensorTypeRotary:
                    if scale_shift >= 360:
                        scale_shift = scale_shift % 360
                    if scale_shift > 180:
                        scale_shift = scale_shift - 360
                return up_down, scale_shift
            except Exception, e:
                PyTango.Except.throw_exception("Get Home Position failed", "Unexpected value returned", "read_scale()")


    def read_logical_position(self):
        """ Returns angle in deg or position in mm."""
        if self.SensorTypeRotary:
            # rotary
            cmd_str = "GA"+str(self.ChannelIndex)
            stype_rotary = True
        else:
            # linear
            cmd_str = "GP"+str(self.ChannelIndex)
            stype_rotary = False
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Read Position Failed", self.status_string, "read_logical_position()")
        else:
            # print 'Reply:',resp
            try:
                if stype_rotary:
                    angle = float(str.split(resp,',')[1])
                    rev = float(str.split(resp,',')[2])
                    if rev < 0:
                        angle = angle - 360e6
                    return angle/self.SensorUnitScaleFactor
                else:
                    pos = float(str.split(resp,',')[1])
                    return pos/self.SensorUnitScaleFactor
            except Exception, e:
                PyTango.Except.throw_exception("Read Position Failed", "Unexpected reply format", "read_logical_position()")        
             

    def read_limits(self):
        # self.debug_stream("In read_limits()")
        if self.SensorTypeRotary:
            # rotary
            cmd_str = "GAL"+str(self.ChannelIndex)
        else:
            # linear
            cmd_str = "GPL"+str(self.ChannelIndex)

        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Read Limits Failed", self.status_string, "read_limits()")
        else:
            # print 'Reply:',resp
            try:
                if self.SensorTypeRotary:
                    min_angle = float(str.split(resp, ',')[1])
                    min_rev = float(str.split(resp, ',')[2])
                    if min_rev < 0:
                        min_angle = min_angle - 360e6
                    self.attr_LowLimitPosition_read = min_angle / self.SensorUnitScaleFactor
                    #
                    max_angle = float(str.split(resp, ',')[3])
                    max_rev = float(str.split(resp, ',')[4])
                    if max_rev < 0:
                        max_angle = max_angle - 360e6
                    self.attr_HighLimitPosition_read = max_angle / self.SensorUnitScaleFactor
                else:
                    self.attr_LowLimitPosition_read = float(str.split(resp, ',')[1]) / self.SensorUnitScaleFactor
                    self.attr_HighLimitPosition_read = float(str.split(resp, ',')[2]) / self.SensorUnitScaleFactor
            except Exception, e:
                PyTango.Except.throw_exception("Read Limits Failed", "Unexpected reply format", "read_limits()")
        return self.attr_LowLimitPosition_read, self.attr_HighLimitPosition_read



    #-----------------------------------------------------------------------------
    #    Smaract_mcs read/write attribute methods
    #-----------------------------------------------------------------------------

    def read_Position(self, attr):
        # self.debug_stream("In read_Position()")
        self.attr_Position_read = self.read_logical_position()
        attr.set_value(self.attr_Position_read)
        if self.get_state() == PyTango.DevState.ON:
            if abs(self.attr_PositionBackup_read - self.attr_Position_read) >= 0.001:
                self.autoproxy.PositionBackup = self.attr_Position_read


    def write_Position(self, attr):
        # self.debug_stream("In write_Position()")
        data = attr.get_write_value()

        if self.attr_RequireGo_read :
            self.attr_NextPosition_read = data
            return

        in_val = int(data * self.SensorUnitScaleFactor)
        if self.SensorTypeRotary:
            # rotary
            if math.fabs(in_val) >= 360e6:
                angle = int(math.fmod(in_val,360e6)) # We take into consideration only one rotation left/right.
            else:
                angle = in_val
            if in_val < 0:
                rev = -1
                angle = int(360e6 + angle) # MAA cmd requires positive values: angle should be normalized.
            else:
                rev = 0
            cmd_str = "MAA"+str(self.ChannelIndex)+","+str(angle)+","+str(rev)+","+str(self.hold_time_ms)
        else:
            # linear
            cmd_str = "MPA"+str(self.ChannelIndex)+","+str(in_val)+","+str(self.hold_time_ms)

        tmpPrevPos = self.attr_Position_read
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Write Position Failed", self.status_string, "write_Position()")
        else:
            self.attr_PreviousPosition_read = tmpPrevPos


    def is_Position_allowed(self, req_type):
        return req_type is PyTango.AttReqType.READ_REQ or self.get_state() in (PyTango.DevState.ON,PyTango.DevState.MOVING,)


    def read_Acceleration(self, attr):
        # self.debug_stream("In read_Acceleration()")
        # acceleration - Defines the acceleration in micro-m/s^2 for linear positioners
        # and in milidegrees/s^2 for rotary positioners. The valid range is 0..10,000,000.
        # A value of 0 (default) deactivates the acceleration control feature.
        cmd_str = "GCLA"+str(self.ChannelIndex)
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Read Acceleration Failed", self.status_string, "read_Acceleration()")
        else:
            # print 'Reply:',resp
            try:
                read_acc = float(str.split(resp,',')[1])
                self.attr_Acceleration_read = read_acc / self.SensorUnitScaleFactor
            except Exception, e:
                PyTango.Except.throw_exception("Read Acceleration Failed", "Unexpected reply format", "read_Acceleration()")

        attr.set_value(self.attr_Acceleration_read)


    def write_Acceleration(self, attr):
        # self.debug_stream("In write_Acceleration()")
        data = attr.get_write_value()
        # acceleration - Defines the acceleration in micro-m/s^2 for linear positioners
        # and in milidegrees/s^2 for rotary positioners. The valid range is 0..10,000,000.
        # A value of 0 (default) deactivates the acceleration control feature.
        acc = int(data * self.SensorUnitScaleFactor)
        cmd_str = "SCLA"+str(self.ChannelIndex)+','+str(acc)
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Write Acceleration Failed", self.status_string, "write_Acceleration()")


    def read_Velocity(self, attr):
        # self.debug_stream("In read_Velocity()")
        # Movement speed given in nm/s for linear positioners and in micro-degrees/s for rotary positioners.
        # The valid range is 0 .. 100,000,000.
        # A value of 0 (default) deactivates the speed control feature.
        cmd_str = "GCLS"+str(self.ChannelIndex)
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Read Velocity Failed", self.status_string, "read_Velocity()")
        else:
            # print 'Reply:',resp
            try:
                read_speed = float(str.split(resp,',')[1])
                self.attr_Velocity_read = read_speed / self.SensorUnitScaleFactor
            except Exception, e:
                PyTango.Except.throw_exception("Read Velocity Failed", "Unexpected reply format", "read_Velocity()")

        attr.set_value(self.attr_Velocity_read)


    def write_Velocity(self, attr):
        # self.debug_stream("In write_Velocity()")
        data = attr.get_write_value()
        # Movement speed given in nm/s for linear positioners and in micro-degrees/s for rotary positioners.
        # The valid range is 0 .. 100,000,000.
        # A value of 0 (default) deactivates the speed control feature.
        speed = int(data * self.SensorUnitScaleFactor)
        cmd_str = "SCLS"+str(self.ChannelIndex)+','+str(speed)
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Write Velocity Failed", self.status_string, "write_Velocity()")
        else:
            self.attr_Velocity_read = data


    def read_MotorTitle(self, attr):
        # self.debug_stream("In read_MotorTitle()")
        attr.set_value(self.attr_MotorTitle_read)


    def write_MotorTitle(self, attr):
        # self.debug_stream("In write_MotorTitle()")
        data = attr.get_write_value()
        self.attr_MotorTitle_read = data


    def read_DefinePosition(self, attr):
        # self.debug_stream("In read_DefinePosition()")
        self.attr_DefinePosition_read = self.read_scale()[1]
        attr.set_value(self.attr_DefinePosition_read)


    def write_DefinePosition(self, attr):
        # self.debug_stream("In write_DefinePosition()")
        data = attr.get_write_value()
        # Sending a SP command sets the offset implicitly by shifting the logical
        # scale so that the current position equals the desired value.

        current_pos = self.read_logical_position()
        up_down = self.read_scale()[0]

        in_val = int(data * self.SensorUnitScaleFactor)
        if self.SensorTypeRotary:
            # rotary
            if math.fabs(in_val) >= 360e6:
                angle = int(math.fmod(in_val, 360e6)) # We take into consideration only one rotation left/right.
            else:
                angle = in_val
            if in_val < 0:
                angle = int(360e6 + angle) # SP cmd requires positive values: angle should be normalized.

        cmd_str = "SP"+str(self.ChannelIndex)+","+str(in_val)
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Write Home Position Failed", self.status_string, "write_DefinePosition()")
        else:
            # estimate the scale shift of the reference mark
            direction = 1
            if up_down:
                direction = -1
            ref_pos = data - direction * current_pos
            # store the estimated position into the Home_position attribute.
            self.autoproxy.Home_position = ref_pos


    def is_DefinePosition_allowed(self, req_type):
        return req_type is PyTango.AttReqType.READ_REQ or self.get_state() in (PyTango.DevState.ON,)

    def read_Home_position(self, attr):
        # self.debug_stream("In read_Home_position()")
        attr.set_value(self.attr_Home_position_read)

    def write_Home_position(self, attr):
        # self.debug_stream("In write_Home_position()")
        data = attr.get_write_value()
        self.attr_Home_position_read = data

    def is_Home_position_allowed(self, req_type):
        return req_type is PyTango.AttReqType.READ_REQ or self.get_state() in (PyTango.DevState.ON,)

    def read_StepSize(self, attr):
        # self.debug_stream("In read_StepSize()")
        attr.set_value(self.attr_StepSize_read)

    def write_StepSize(self, attr):
        # self.debug_stream("In write_StepSize()")
        data = attr.get_write_value()
        self.attr_StepSize_read = data


    def read_InvertUpDown(self, attr):
        # self.debug_stream("In read_InvertUpDown()")
        # 
        # cmd_str = "GSC"+str(self.ChannelIndex)
        # resp = self.socket_write_and_read(cmd_str)
        # if resp == 0:
        #     self.debug_stream(self.attr_Error_read)
        #     PyTango.Except.throw_exception("Get Inverted Failed", self.status_string, "read_InvertUpDown()")
        # else:
        #     # print 'Reply:',resp
        #     try:
        #         inverted = int(str.split(resp, ',')[2])
        #         if inverted:
        #             self.attr_InvertUpDown_read = True
        #         else:
        #             self.attr_InvertUpDown_read = False
        #     except Exception, e:
        #         PyTango.Except.throw_exception("Get Inverted Failed", "Unexpected value returned", "read_InvertUpDown()")
        # 
        self.attr_InvertUpDown_read = self.read_scale()[0]
        attr.set_value(self.attr_InvertUpDown_read)


    def write_InvertUpDown(self, attr):
        # self.debug_stream("In write_InvertUpDown()")
        data = attr.get_write_value()

        # Sending a SSC command sets the offset explicitly and the current position
        # will have a value that reflects the new scale shift and direction.

        # The offset and inversion value is stored in non-volatile memory.
        # Once it is configured you only need to send a FRM to restore your settings
        # on future power-ups.

        # Parameters:
        # channelIndex - Selects the channel of the system. The index is zero based.
        # scale – Sets the desired scale shift relative to the physical scale of the positioner.
        #   The value is given in nano meters for linear positioners
        #   and in micro degrees for rotary positioners.
        # inverted – Sets the scale inversion. Must be either 0 (disabled) or 1 (enabled).

        if data:
            inverted = 1
        else:
            inverted = 0

        cmd_str = "GSC"+str(self.ChannelIndex)
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Get Home Position failed", self.status_string, "write_InvertUpDown()")
        else:
            # print 'Reply:',resp
            try:
                logical_shift = int(str.split(resp, ',')[1])
            except Exception, e:
                PyTango.Except.throw_exception("Get Home Position failed", "Unexpected value returned", "write_InvertUpDown()")

        cmd_str = "SSC"+str(self.ChannelIndex)+',' + str(logical_shift) + ','+str(inverted)
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Set Inverted Failed", self.status_string, "write_InvertUpDown()")
        else:
            self.attr_InvertUpDown_read = data


    def is_InvertUpDown_allowed(self, req_type):
        return req_type is PyTango.AttReqType.READ_REQ or self.get_state() in (PyTango.DevState.ON,)


    def read_ClosedLoop(self, attr):
        # self.debug_stream("In read_ClosedLoop()")
        if self.ChannelIndex == 8:
            self.attr_ClosedLoop_read = False
        else:
            self.attr_ClosedLoop_read = True
        attr.set_value(self.attr_ClosedLoop_read)


    def read_NextPosition(self, attr):
        # self.debug_stream("In read_NextPosition()")
        attr.set_value(self.attr_NextPosition_read)


    def write_NextPosition(self, attr):
        # self.debug_stream("In write_NextPosition()")
        data = attr.get_write_value()
        self.attr_NextPosition_read = data
        # return
        # The following matches the logic of other motor devices (Micos, etc)
        if not self.attr_RequireGo_read:
            in_val = int(data * self.SensorUnitScaleFactor)
            if self.SensorTypeRotary:
                # rotary
                if math.fabs(in_val) >= 360e6:
                    angle = int(math.fmod(in_val, 360e6))
                    # We take into consideration only one rotation left/right.
                else:
                    angle = in_val
                if in_val < 0:
                    rev = -1
                    angle = int(360e6 + angle) # MAA cmd requires positive values: angle should be normalized.
                else:
                    rev = 0
                cmd_str = "MAA"+str(self.ChannelIndex)+","+str(angle)+","+str(rev)+","+str(self.hold_time_ms)
            else:
                # linear
                cmd_str = "MPA"+str(self.ChannelIndex)+","+str(in_val)+","+str(self.hold_time_ms)
            tmpPrevPos = self.attr_Position_read
            resp = self.socket_write_and_read(cmd_str)
            if resp == 0:
                self.debug_stream(self.attr_Error_read)
                PyTango.Except.throw_exception("Write NextPosition Failed", self.status_string, "write_NextPosition()")
            else:
                self.attr_PreviousPosition_read = tmpPrevPos

    def read_PreviousPosition(self, attr):
        # self.debug_stream("In read_PreviousPosition()")
        attr.set_value(self.attr_PreviousPosition_read)


    def read_Error(self, attr):
        # self.debug_stream("In read_Error()")
        if self.error_msg_timer > 0 and time.time() - self.error_msg_timer > 30:
            self.attr_Error_read = ''
            self.error_msg_timer = 0.0
        attr.set_value(self.attr_Error_read)


    def read_RequireGo(self, attr):
        # self.debug_stream("In read_RequireGo()")
        attr.set_value(self.attr_RequireGo_read)


    def write_RequireGo(self, attr):
        # self.debug_stream("In write_RequireGo()")
        data = attr.get_write_value()
        self.attr_RequireGo_read = data

    def read_HighLimitPosition(self, attr):
        # self.debug_stream("In read_HighLimitPosition()")
        self.attr_LowLimitPosition_read, self.attr_HighLimitPosition_read = self.read_limits()
        attr.set_value(self.attr_HighLimitPosition_read)


    def write_HighLimitPosition(self, attr):
        # self.debug_stream("In write_HighLimitPosition()")
        data = attr.get_write_value()
        self.attr_LowLimitPosition_read, self.attr_HighLimitPosition_read = self.read_limits()
        in_val = int(data * self.SensorUnitScaleFactor)
        low_lim = int(self.attr_LowLimitPosition_read * self.SensorUnitScaleFactor)

        if self.SensorTypeRotary:
            # rotary
            if math.fabs(in_val) >= 360e6:
                in_val = int(math.fmod(in_val, 360e6))
                # We take into consideration only one rotation left/right.
            if in_val < low_lim:
                self.warn_stream('Set the low limit first.')
                PyTango.Except.throw_exception("Cannot write High Limit Position", "Low limit not set or higher then high limit.", "write_HighLimitPosition()")
                return
            if in_val < 0:
                max_rev = -1
                max_angle = int(360e6 + in_val) # SAL cmd requires positive values: angle should be normalized.
            else:
                max_rev = 0
                max_angle = in_val
            if low_lim < 0:
                min_angle = int(360e6 + low_lim)
                min_rev = -1
            else:
                min_angle = low_lim
                min_rev = 0

            cmd_str = "SAL"+str(self.ChannelIndex)+","+str(min_angle)+","+\
            str(min_rev)+","+str(max_angle)+","+str(max_rev)
        else:
            if in_val < low_lim:
                self.warn_stream('Set the low limit first.')
                PyTango.Except.throw_exception("Cannot write High Limit Position", "Low limit not set or higher then high limit.", "write_HighLimitPosition()")
                return
            cmd_str = "SPL"+str(self.ChannelIndex)+","+str(low_lim)+","+str(in_val)
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Write High Limit Position Failed", self.status_string, "write_HighLimitPosition()")
        else:
            self.attr_HighLimitPosition_read = data


    def is_HighLimitPosition_allowed(self, req_type):
        return req_type is PyTango.AttReqType.READ_REQ or self.get_state() in (PyTango.DevState.ON,)


    def read_LowLimitPosition(self, attr):
        # self.debug_stream("In read_LowLimitPosition()")
        self.attr_LowLimitPosition_read, self.attr_HighLimitPosition_read = self.read_limits()
        attr.set_value(self.attr_LowLimitPosition_read)


    def write_LowLimitPosition(self, attr):
        # self.debug_stream("In write_LowLimitPosition()")
        data = attr.get_write_value()
        self.attr_LowLimitPosition_read, self.attr_HighLimitPosition_read = self.read_limits()
        in_val = int(data * self.SensorUnitScaleFactor)
        high_lim = int(self.attr_HighLimitPosition_read * self.SensorUnitScaleFactor)

        if self.SensorTypeRotary:
            # rotary
            if math.fabs(in_val) >= 360e6:
                in_val = int(math.fmod(in_val,360e6)) # We take into consideration only one rotation left/right.
            if in_val > high_lim:
                self.warn_stream('Set the high limit first.')
                PyTango.Except.throw_exception("Cannot write Low Limit Position", "High limit not set or lower then low limit.", "write_LowLimitPosition()")
                return
            if in_val < 0:
                min_rev = -1
                min_angle = int(360e6 + in_val) # SAL cmd requires positive values: angle should be normalized.
            else:
                min_rev = 0
                min_angle = in_val
            if high_lim < 0:
                max_angle = int(360e6 + high_lim)
                max_rev = -1
            else:
                max_angle = high_lim
                max_rev = 0

            cmd_str = "SAL"+str(self.ChannelIndex)+","+str(min_angle)+","+str(min_rev)+","+str(max_angle)+","+str(max_rev)
        else:
            if in_val > high_lim:
                self.warn_stream('Set the high limit first.')
                PyTango.Except.throw_exception("Cannot write Low Limit Position", "High limit not set or lower then low limit.", "write_LowLimitPosition()")
                return
            cmd_str = "SPL"+str(self.ChannelIndex)+","+str(in_val)+","+str(high_lim)

        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Write Low Limit Position Failed", self.status_string, "write_LowLimitPosition()")
        else:
            self.attr_LowLimitPosition_read = data


    def is_LowLimitPosition_allowed(self, req_type):
        return req_type is PyTango.AttReqType.READ_REQ or self.get_state() in (PyTango.DevState.ON,)


    def read_PositionBackup(self, attr):
        # self.debug_stream("In read_PositionBackup()")
        attr.set_value(self.attr_PositionBackup_read)


    def write_PositionBackup(self, attr):
        # self.debug_stream("In write_PositionBackup()")
        data = attr.get_write_value()
        self.attr_PositionBackup_read = data


    def read_SensorEnabled(self, attr):
        # self.debug_stream("In read_SensorEnabled()")
        cmd_str = "GSE"
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Read Sensor Enabled Failed", self.status_string, "read_SensorEnabled()")
        else:
            # print 'GP:',resp
            try:
                self.attr_SensorEnabled_read = int(resp[-1])
            except Exception, e:
                PyTango.Except.throw_exception("Read Sensor Enabled Failed", "Unexpected reply format", "read_SensorEnabled()")

        attr.set_value(self.attr_SensorEnabled_read)


    def write_SensorEnabled(self, attr):
        # self.debug_stream("In write_SensorEnabled()")
        data = attr.get_write_value()
        cmd_str = "SSE"+str(data)
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Write Sensor Enabled Failed", self.status_string, "write_SensorEnabled()")


    def read_SensorStatus(self, attr):
        # self.debug_stream("In read_SensorStatus()")
        cmd_str = "GS"+str(self.ChannelIndex)
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Read Sensor Status Failed", self.status_string, "read_SensorStatus()")
        else:
            # print 'GP:',resp
            try:
                status_code = int(str.split(resp,',')[1])
                self.attr_SensorStatus_read = self.get_movement_status(status_code)
            except Exception, e:
                PyTango.Except.throw_exception("Read Sensor Status Failed", "Unexpected reply format", "read_SensorStatus()")

        attr.set_value(self.attr_SensorStatus_read)


    def read_PhysicalPositionKnown(self, attr):
        # self.debug_stream("In read_PhysicalPositionKnown()")
        # Get physical position known (1 - yes, 0 - no)
        cmd = "GPPK" + str(self.ChannelIndex)
        resp = self.socket_write_and_read(cmd)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            self.set_state(PyTango.DevState.FAULT)
            PyTango.Except.throw_exception("Get Physical Position Known failed", self.status_string, "read_PhysicalPositionKnown()")
        else:
            if int(str.split(resp, ',')[1]):
                # Expected result is PPK,<val> 1 = known, 0 = unknown
                self.attr_PhysicalPositionKnown_read = True
                self.set_state(PyTango.DevState.ON)
                self.status_string = "Device ready."
            else:
                self.attr_PhysicalPositionKnown_read = False
                self.set_state(PyTango.DevState.ALARM)
                self.status_string = "Physical Position Unknown!"

        attr.set_value(self.attr_PhysicalPositionKnown_read)


    def read_BiDirReferencing(self, attr):
        # self.debug_stream("In read_BiDirReferencing()")
        attr.set_value(self.attr_BiDirReferencing_read)


    def write_BiDirReferencing(self, attr):
        # self.debug_stream("In write_BiDirReferencing()")
        data = attr.get_write_value()
        self.attr_BiDirReferencing_read = data


    def read_attr_hardware(self, data):
        # self.debug_stream("In read_attr_hardware()")
        return


    #-----------------------------------------------------------------------------
    #    Smaract_mcs command methods
    #-----------------------------------------------------------------------------

    def On(self):
        """
        :param :
        :type: PyTango.DevVoid
        :return:
        :rtype: PyTango.DevVoid """
        #self.debug_stream("In On()")
        return

    def Off(self):
        """
        :param :
        :type: PyTango.DevVoid
        :return:
        :rtype: PyTango.DevVoid """
        # self.debug_stream("In Off()")
        self.set_state(PyTango.DevState.OFF)

    def Abort(self):
        """
        :param :
        :type: PyTango.DevVoid
        :return:
        :rtype: PyTango.DevVoid """
        # self.debug_stream("In Abort()")
        # Same as STOP?
        cmd_str = "S"+str(self.ChannelIndex)
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Cmd Abort Failed", self.status_string, "Abort()")
        else:
            self.status_string = "Abort (Stop) executed."
        

    def CalibrateSensor(self):
        """ 
        Runs calibrate sensor mark on the controller.
        :param :
        :type: PyTango.DevVoid
        :return:
        :rtype: PyTango.DevVoid """
        # self.debug_stream("In CalibrateSensor()")
        cmd = "CS"+str(self.ChannelIndex)
        if self.socket_write_and_read(cmd) == 0:
            self.debug_stream(self.attr_Error_read)
            self.set_state(PyTango.DevState.FAULT)
            PyTango.Except.throw_exception("Calibrate Sensor failed", self.status_string, "CalibrateSensor()")
        else:
            self.status_string = "Calibrate Sensor executed."
            self.debug_stream("CalibrateSensor(): Calibrate Sensor executed.\n")
            

    def Go(self):
        """
        :param :
        :type: PyTango.DevVoid
        :return:
        :rtype: PyTango.DevVoid """
        # self.debug_stream("In Go()")
        if self.attr_NextPosition_read == self.attr_Position_read:
            return
        in_val = int(self.attr_NextPosition_read * self.SensorUnitScaleFactor)
        if self.SensorTypeRotary:
            # rotary
            if math.fabs(in_val) >= 360e6:
                angle = int(math.fmod(in_val,360e6)) # We take into consideration only one rotation left/right.
            else:
                angle = in_val
            if in_val < 0:
                rev = -1
                angle = int(360e6 + angle) # MAA cmd requires positive values: angle should be normalized.
            else:
                rev = 0
            cmd_str = "MAA"+str(self.ChannelIndex)+","+str(angle)+","+str(rev)+","+str(self.hold_time_ms)
        else:
            # linear
            cmd_str = "MPA"+str(self.ChannelIndex)+","+str(in_val)+","+str(self.hold_time_ms)

        tmpPrevPos = self.attr_Position_read
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Cmd Go Failed", self.status_string, "Go()")
        else:
            self.attr_PreviousPosition_read = tmpPrevPos


    def is_Go_allowed(self):
        return self.get_state() in (PyTango.DevState.ON,)     


    def StepUp(self):
        """
        :param :
        :type: PyTango.DevVoid
        :return:
        :rtype: PyTango.DevVoid """
        # self.debug_stream("In StepUp()")
        if self.attr_StepSize_read == 0:
            self.status_string = 'Step size is not set.'
            self.set_status(self.status_string)
            return
        if self.attr_RequireGo_read :
            curr_pos = self.autoproxy.Position
            self.attr_NextPosition_read = curr_pos + self.attr_StepSize_read
            return
        step = int(self.attr_StepSize_read * self.SensorUnitScaleFactor)
        if self.SensorTypeRotary:
            # rotary - no revolution difference
            cmd_str = "MAR"+str(self.ChannelIndex)+","+str(step)+",0,"+str(self.hold_time_ms)
        else:
            # linear
            cmd_str = "MPR"+str(self.ChannelIndex)+","+str(step)+","+str(self.hold_time_ms)
        tmpPrevPos = self.attr_Position_read
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Cmd Step Up Failed", self.status_string, "StepUp()")
        else:
            self.attr_PreviousPosition_read = tmpPrevPos


    def is_StepUp_allowed(self):
        return self.get_state() in (PyTango.DevState.ON,)     


    def StepDown(self):
        """
        :param :
        :type: PyTango.DevVoid
        :return:
        :rtype: PyTango.DevVoid """
        # self.debug_stream("In StepDown()")
        if self.attr_StepSize_read == 0:
            self.status_string = 'Step size is not set.'
            self.set_status(self.status_string)
            return
        if self.attr_RequireGo_read :
            curr_pos = self.autoproxy.Position
            self.attr_NextPosition_read = curr_pos - self.attr_StepSize_read
            return
        step = -1 * int(self.attr_StepSize_read * self.SensorUnitScaleFactor)
        if self.SensorTypeRotary:
            # rotary - no revolution difference
            cmd_str = "MAR"+str(self.ChannelIndex)+","+str(step)+",0,"+str(self.hold_time_ms)
        else:
            # linear
            cmd_str = "MPR"+str(self.ChannelIndex)+","+str(step)+","+str(self.hold_time_ms)

        tmpPrevPos = self.attr_Position_read
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Cmd Step Up Failed", self.status_string, "StepUp()")
        else:
            self.attr_PreviousPosition_read = tmpPrevPos


    def is_StepDown_allowed(self):
        return self.get_state() in (PyTango.DevState.ON,)         


    def AdminState(self, argin):
        """
        :param argin:
        :type: PyTango.DevString
        :return:
        :rtype: PyTango.DevVoid """
        # self.debug_stream("In AdminState()")
        if argin == 'kik':
            self.debug_stream('Admin rights granted.')
        else:
            self.warn_stream('Admin rights refused, incorrect password.')
            self.status_string = "Incorrect password"
            PyTango.Except.throw_exception("Admin access denied.", self.status_string, "AdminState()")


    def GoHome(self):
        """ Find the reference mark and set the scale shift.
        :param :
        :type: PyTango.DevVoid
        :return:
        :rtype: PyTango.DevVoid """
        # self.debug_stream("In GoHome()")
        self.debug_stream("In GoHome() == FindReferenceMark() + set scale shift as it was.")
        self.FindReferenceMark()
        time.sleep(1.5)
        # Set the physical 0 position to Home_position value. (Scale Shift)
        db = PyTango.Database()
        dev_attr_props = db.get_device_attribute_property(self.get_name(), "Home_position")
        try:
            last_home_pos = float(dev_attr_props['Home_position']['__value'][0])
        except:
            last_home_pos = 0.0
        if last_home_pos != 0:
            in_val = int(last_home_pos * self.SensorUnitScaleFactor)
            if self.SensorTypeRotary:
                # rotary
                if math.fabs(in_val) >= 360e6:
                    angle = int(math.fmod(in_val, 360e6)) # We take into consideration only one rotation left/right.
                else:
                    angle = in_val
                if in_val < 0:
                    angle = int(360e6 + angle) # SP cmd requires positive values: angle should be normalized.

            cmd_str = "SP"+str(self.ChannelIndex)+","+str(in_val)
            resp = self.socket_write_and_read(cmd_str)
            if resp == 0:
                self.debug_stream(self.attr_Error_read)
                PyTango.Except.throw_exception("Define Position Failed", self.status_string, "GoHome()")
            else:
                self.info_stream("Reference mark found, scale shift set to " + str(last_home_pos))
        else:
            self.info_stream("Reference mark found, scale shift 0.")


    def GoToHighLimit(self):
        """
        :param :
        :type: PyTango.DevVoid
        :return:
        :rtype: PyTango.DevVoid """
        # self.debug_stream("In GoToHighLimit()")
        if self.attr_HighLimitPosition_read == self.attr_Position_read:
            return
        in_val = int(self.attr_HighLimitPosition_read * self.SensorUnitScaleFactor)
        if self.SensorTypeRotary:
            # rotary
            if math.fabs(in_val) >= 360e6:
                angle = int(math.fmod(in_val,360e6)) # We take into consideration only one rotation left/right.
            else:
                angle = in_val
            if in_val < 0:
                rev = -1
                angle = int(360e6 + angle) # MAA cmd requires positive values: angle should be normalized.
            else:
                rev = 0
            cmd_str = "MAA"+str(self.ChannelIndex)+","+str(angle)+","+str(rev)+","+str(self.hold_time_ms)
        else:
            # linear
            cmd_str = "MPA"+str(self.ChannelIndex)+","+str(in_val)+","+str(self.hold_time_ms)
        tmpPrevPos = self.attr_Position_read
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Cmd Go To High Limit Failed", self.status_string, "GoToHighLimit()")
        else:
            self.attr_PreviousPosition_read = tmpPrevPos


    def is_GoToHighLimit_allowed(self):
        return self.get_state() in (PyTango.DevState.ON,)         


    def GoToLowLimit(self):
        """
        :param :
        :type: PyTango.DevVoid
        :return:
        :rtype: PyTango.DevVoid """
        # self.debug_stream("In GoToLowLimit()")
        if self.attr_LowLimitPosition_read == self.attr_Position_read:
            return
        in_val = int(self.attr_LowLimitPosition_read * self.SensorUnitScaleFactor)
        if self.SensorTypeRotary:
            # rotary
            if math.fabs(in_val) >= 360e6:
                angle = int(math.fmod(in_val,360e6)) # We take into consideration only one rotation left/right.
            else:
                angle = in_val
            if in_val < 0:
                rev = -1
                angle = int(360e6 + angle) # MAA cmd requires positive values: angle should be normalized.
            else:
                rev = 0
            cmd_str = "MAA"+str(self.ChannelIndex)+","+str(angle)+","+str(rev)+","+str(self.hold_time_ms)
        else:
            # linear
            cmd_str = "MPA"+str(self.ChannelIndex)+","+str(in_val)+","+str(self.hold_time_ms)
        tmpPrevPos = self.attr_Position_read
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Cmd Go To Low Limit Failed", self.status_string, "GoToLowLimit()")
        else:
            self.attr_PreviousPosition_read = tmpPrevPos

    def is_GoToLowLimit_allowed(self):
        return self.get_state() in (PyTango.DevState.ON,)           


    def GetParameters(self):
        """
        :param :
        :type: PyTango.DevVoid
        :return:
        :rtype: PyTango.DevVoid """
        # self.debug_stream("In GetParameters()")
        self.attr_Position_read = self.autoproxy.Position
        self.attr_NextPosition_read = self.autoproxy.NextPosition
        self.attr_PreviousPosition_read = self.autoproxy.PreviousPosition
        self.attr_Velocity_read = self.autoproxy.Velocity
        self.attr_Acceleration_read = self.autoproxy.Acceleration
        self.attr_StepSize_read = self.autoproxy.StepSize
        self.attr_HighLimitPosition_read = self.autoproxy.HighLimitPosition
        self.attr_LowLimitPosition_read = self.autoproxy.LowLimitPosition
        self.attr_InvertUpDown_read = self.autoproxy.InvertUpDown
        self.attr_DefinePosition_read = self.autoproxy.DefinePosition
        self.attr_Home_position_read = self.autoproxy.Home_position


    def is_GetParameters_allowed(self):
        return self.get_state() in (PyTango.DevState.ON,)    


    def Stop(self):
        """
        :param :
        :type: PyTango.DevVoid
        :return:
        :rtype: PyTango.DevVoid """
        # self.debug_stream("In Stop()")
        cmd_str = "S"+str(self.ChannelIndex)
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Cmd Stop Failed", self.status_string, "Stop()")


    def Reset(self):
        """
        :param :
        :type: PyTango.DevVoid
        :return:
        :rtype: PyTango.DevVoid """
        # self.debug_stream("In Reset()")
        self.debug_stream("In Reset() == Init()")
        self.autoproxy.Init()


    def FindReferenceMark(self):
        """ Find the reference mark of the positioner (tipically near the middle of the travel range)
        :param :
        :type: PyTango.DevVoid
        :return:
        :rtype: PyTango.DevVoid """
        # self.debug_stream("In FindReferenceMark()")
        # To decide the direction in which to search for the reference mark, we will check
        # the last setting of the PositionBackup attribute (set automatically at each )
        db = PyTango.Database()
        dev_attr_props = db.get_device_attribute_property(self.get_name(), "PositionBackup")
        # print dev_attr_props
        try:
            last_known_pos = float(dev_attr_props['PositionBackup']['__value'][0])
        except:
            last_known_pos = 0.0
        self.debug_stream('Last Known Position >>> ' +str(last_known_pos))
        #
        dev_attr_props = db.get_device_attribute_property(self.get_name(), "Home_position")
        # print dev_attr_props
        try:
            last_home_pos = float(dev_attr_props['Home_position']['__value'][0])
        except:
            last_home_pos = 0.0
        self.debug_stream('Last Known Position >>> ' +str(last_home_pos))
        #
        if self.attr_InvertUpDown_read:
            physical_position = last_home_pos - last_known_pos
        else:
            physical_position = last_home_pos + last_known_pos
        #
        self.debug_stream('Physical position calculated >>> ' +str(physical_position))
        # Find Reference Mark (FRM) Params:
        #  channelIndex - Selects the channel of the system. The index is zero based.
        #  direction - 0 (forward only), 1 (backward only), 2 (forward backward), 3 (backward forward), 
        #              4 (forward with abort on endstop), 5 (backward with abort on endstop), 
        #              6 (forward backward with abort on endstop) or 7 (backward forward with abort on endstop).
        #  holdTime - Specifies how long (in milliseconds) the position is actively held after reaching the reference mark.
        #  autoZero - Must be either 0 (no auto zero) or 1 (auto zero).
        #  If the auto zero flag is set, the current position resp. angle is set to zero after the reference position has been reached.
        if physical_position <= 0:
            if self.attr_BiDirReferencing_read:
                self.safe_direction = 2
            else:
                self.safe_direction = 0
        else:
            if self.attr_BiDirReferencing_read:
                self.safe_direction = 3
            else:
                self.safe_direction = 1
        self.set_state(PyTango.DevState.INIT)
        cmd = "FRM" + str(self.ChannelIndex) + "," + str(self.safe_direction) + ",1000,1"
        if self.socket_write_and_read(cmd) == 0:
            self.debug_stream(self.attr_Error_read)
            self.set_state(PyTango.DevState.FAULT)
            PyTango.Except.throw_exception("Find Reference Mark failed", self.status_string, "FindReferenceMark()")
        else:
            time.sleep(0.5) # Give the controller some time to execute.
            print "FindReferenceMark(): Find Refernce Mark executed."
            self.set_state(PyTango.DevState.ON)
            self.attr_PhysicalPositionKnown_read = self.autoproxy.PhysicalPositionKnown


    def ForceFindRefMark(self, argin):
        """ Find the reference mark of the positioner (tipically near the middle of the travel range)
        :param :
        :type: PyTango.DevShort
            Must be either:
            0 (forward only),
            1 (backward only),
            2 (forward backward),
            3 (backward forward),
            4 (forward with abort on endstop),
            5 (backward with abort on endstop),
            6 (forward backward with abort on endstop),
            7 (backward forward with abort on endstop).
        :return:
        :rtype: PyTango.DevVoid """
        # self.debug_stream("In ForceFindRefMark()")

        if not 0 <= argin <= 7:
            PyTango.Except.throw_exception("Unable to execute FRM", \
                "Input value out of range [0-7]", "ForceFindRefMark()")

        cmd = "FRM" + str(self.ChannelIndex) + "," + str(argin) + ",1000,1"
        if self.socket_write_and_read(cmd) == 0:
            self.debug_stream(self.attr_Error_read)
            self.set_state(PyTango.DevState.FAULT)
            PyTango.Except.throw_exception("Find Reference Mark failed", self.status_string, "ForceFindRefMark()")
        else:
            time.sleep(0.5) # Give the controller some time to execute.
            print "FindReferenceMark(): Find Refernce Mark executed."
            self.set_state(PyTango.DevState.ON)
            # force read PhysicalPositionKnown
            self.attr_PhysicalPositionKnown_read = self.autoproxy.PhysicalPositionKnown


    def write_and_read_command(self, argin):
        """ This command will be used to iteract with the motor in a console like maner.
        It must return either an empty string, or a string with the answer.
        It is advisable not to add a new line at the end of the answer.

        :param argin:
        :type: PyTango.DevString
        :return:
        :rtype: PyTango.DevString """
        # self.debug_stream("In write_and_read_command()")
        argout = ''
        argout = self.socket_write_and_read(argin)
        if argout == 0:
            self.debug_stream(self.attr_Error_read)
            PyTango.Except.throw_exception("Write_and_read Cmd Failed", self.status_string, "write_and_read_command()")
        return argout


class Smaract_mcsClass(PyTango.DeviceClass):
    #--------- Add you global class variables here --------------------------

    def dyn_attr(self, dev_list):
        """Invoked to create dynamic attributes for the given devices.
        Default implementation calls
        :meth:`Smaract_mcs.initialize_dynamic_attributes` for each device

        :param dev_list: list of devices
        :type dev_list: :class:`PyTango.DeviceImpl`"""

        # Have none
        return

        for dev in dev_list:
            try:
                dev.initialize_dynamic_attributes()
            except:
                import traceback
                dev.warn_stream("Failed to initialize dynamic attributes")
                dev.debug_stream("Details: " + traceback.format_exc())


    #    Class Properties
    class_property_list = {
        }


    #    Device Properties
    device_property_list = {
        'SocketDevice':
            [PyTango.DevString,
            "Tango Socket Device that connects to a physical device.",
            []],
        'ChannelIndex':
            [PyTango.DevShort,
            "Integer number of MCS controller channel (1-12)",
            []],
        'SensorType':
            [PyTango.DevShort,
            "Type of sensor attached to the MCS controller channel.",
            []],
        }


    #    Command definitions
    cmd_list = {
        'On':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'Off':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'Abort':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'CalibrateSensor':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'Go':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'StepUp':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'StepDown':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'AdminState':
            [[PyTango.DevString, "none"],
            [PyTango.DevVoid, "none"]],
        'GoHome':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'GoToHighLimit':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'GoToLowLimit':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'GetParameters':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'write_and_read_command':
            [[PyTango.DevString, "none"],
            [PyTango.DevString, "none"]],
        'Stop':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'Reset':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'FindReferenceMark':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'ForceFindRefMark':
            [[PyTango.DevShort, "Direction: 0-7"],
            [PyTango.DevVoid, "none"]],
        }


    #    Attribute definitions
    attr_list = {
        'Position':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "Logical position in mm for linear sensors and \
                in degrees for rotary sensors. Depends on physical position, \
                scale shift and inverted axis direction."
            }],
        'Acceleration':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'max value': "10.0",
                'min value': "0.0",
                'description': "Defines the acceleration in mm/s^2 for linear \
                positioners and in degrees/s^2 for rotary positioners. \
                The valid range is 0..10. A value of 0 (default) deactivates \
                the acceleration control feature.",
            }],
        'Velocity':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'max value': "100.0",
                'min value': "0.0",
                'description': "Defines the speed in mm/s for linear positioners \
                and in degrees/s for rotary positioners. The valid range is 0..100. \
                A value of 0 (default) deactivates the speed control feature.",
            }],
        'MotorTitle':
            [[PyTango.DevString,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "Motor's given name for the GUI's title label.",
                'Memorized':"true"
            }],
        'DefinePosition':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "Defines the scale shift. Sets the current \
                physial position to the logical value provided in input. \
                In output, returns the scale shift value.",
            }],
        'Home_position':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "Evaluated reference mark position \
                with the current scale.",
                'Memorized':"true"
            }],
        'StepSize':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "Step size in mm for linear sensors \
                and in degrees for rotary sensors.",
                'Memorized':"true"
            }],
        'InvertUpDown':
            [[PyTango.DevBoolean,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "If true, axis direction is inverted.",
            }],
        'ClosedLoop':
            [[PyTango.DevBoolean,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'description': "Closed loop means that \
                the sensor has the encoder attached.",
            }],
        'NextPosition':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "Next position in mm for linear sensors and \
                in degrees for rotary sensors. Requires go() command to move.",
            }],
        'PreviousPosition':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'description': "Previous position held, in mm for \
                linear sensors and in degrees for rotary sensors.",
            }],
        'Error':
            [[PyTango.DevString,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'description': "Description of the last error occured.",
            }],
        'RequireGo':
            [[PyTango.DevBoolean,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "If selected, write Position is disabled. \
                Instead, the next position attribute must be set and \
                Go command executed.",
                'Memorized':"true"
            }],
        'HighLimitPosition':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "High limit position in mm for linear sensors \
                and in degrees for rotary sensors.",
            }],
        'LowLimitPosition':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "Low limit position in mm for linear sensors \
                and in degrees for rotary sensors.",
            }],
        'PositionBackup':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "Last known position in mm for linear sensors \
                and in degrees for rotary sensors.",
                'Memorized':"true"
            }],
        'SensorEnabled':
            [[PyTango.DevShort,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'max value': "2",
                'min value': "0",
                'description': "Selects the mode. Must be either \
                0 (disabled), 1 (enabled) or 2 (powersave).",
            }],
        'SensorStatus':
            [[PyTango.DevString,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'description': "Current status of the positioner.",
            }],
        'PhysicalPositionKnown':
            [[PyTango.DevBoolean,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'description': "Checks if the sensor knows its physical \
                position. If not, reference mark must be searched for.",
            }],
        'BiDirReferencing':
            [[PyTango.DevBoolean,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "If true, search for reference mark in both directions.",
                'Memorized':"true"
            }],
        }


def main():
    try:
        py = PyTango.Util(sys.argv)
        py.add_class(Smaract_mcsClass,Smaract_mcs,'Smaract_mcs')

        U = PyTango.Util.instance()
        U.server_init()
        U.server_run()

    except PyTango.DevFailed,e:
        print '-------> Received a DevFailed exception:',e
    except Exception,e:
        print '-------> An unforeseen exception occured....',e

if __name__ == '__main__':
    main()
